from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers
from rest_framework.reverse import reverse

from .models import Source


class SourceSerializer(serializers.HyperlinkedModelSerializer):
    actions = serializers.SerializerMethodField()

    class Meta:
        extra_kwargs = {
            'url': {
                'lookup_field': 'pk', 'lookup_url_kwarg': 'source_id',
                'view_name': 'rest_api:source-detail'
            },
        }
        fields = (
            'actions', 'backend_data', 'backend_path', 'enabled', 'id',
            'label', 'url'
        )
        model = Source
        read_only_fields = ('actions', 'id', 'url')

    def get_actions(self, instance):
        result = []

        for action in instance.get_actions():
            result.append(
                {
                    'name': action.name,
                    'arguments': action.arguments,
                    'url': reverse(
                        viewname='rest_api:source-action', kwargs={
                            'source_id': instance.pk,
                            'action_name': action.name
                        }, request=self.context['request'],
                        format=self.context['format']
                    )
                }
            )

        return result


class SourceBackendActionSerializer(serializers.Serializer):
    arguments = serializers.JSONField(
        help_text=_(
            'Optional arguments for the action. Must be JSON formatted.'
        ), label=_('Arguments'), required=False
    )



from django.db.utils import IntegrityError

from mayan.apps.documents.tests.base import (
    GenericDocumentTestCase, GenericTransactionDocumentTestCase
)
from mayan.apps.documents.tests.literals import (
    TEST_DOCUMENT_DESCRIPTION, TEST_DOCUMENT_DESCRIPTION_EDITED,
    TEST_DOCUMENT_LABEL_EDITED
)
from mayan.apps.metadata.models import MetadataType, DocumentTypeMetadataType

from ..models import (
    IndexInstance, IndexInstanceNode, IndexTemplate, IndexTemplateNode
)

from .literals import (
    TEST_INDEX_TEMPLATE_DOCUMENT_DESCRIPTION_EXPRESSION,
    TEST_INDEX_TEMPLATE_DOCUMENT_LABEL_EXPRESSION,
    TEST_INDEX_TEMPLATE_DOCUMENT_TYPE_EXPRESSION,
    TEST_INDEX_TEMPLATE_METADATA_EXPRESSION, TEST_METADATA_TYPE_LABEL,
    TEST_METADATA_TYPE_NAME
)
from .mixins import IndexTemplateTestMixin


class IndexTemplateTestCase(IndexTemplateTestMixin, GenericDocumentTestCase):
    auto_upload_test_document = False

    def test_method_get_absolute_url(self):
        self._create_test_index_template()
        self.assertTrue(self.test_index_template.get_absolute_url())


class IndexInstanceTestCase(IndexTemplateTestMixin, GenericDocumentTestCase):
    auto_upload_test_document = False
    auto_create_test_index_template_node = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_date_based_index(self):
        level_year = self.test_index_template.index_template_nodes.create(
            parent=self.test_index_template.index_template_root_node,
            expression='{{ document.datetime_created|date:"Y" }}',
            link_documents=False
        )

        self.test_index_template.index_template_nodes.create(
            parent=level_year,
            expression='{{ document.datetime_created|date:"m" }}',
            link_documents=True
        )
        # Index the document created by default.
        IndexTemplate.objects.rebuild()

        self.test_document.delete()

        # Uploading a new should not trigger an error.
        self._upload_test_document()

        self.assertEqual(
            list(IndexInstanceNode.objects.values_list('value', flat=True)),
            [
                '', str(self.test_document.datetime_created.year),
                '{:02}'.format(self.test_document.datetime_created.month)
            ]
        )

        self.assertTrue(
            self.test_document in IndexInstanceNode.objects.last().documents.all()
        )

    def test_dual_level_dual_document_index(self):
        """
        Test creation of an index instance with two first levels with different
        values and two second levels with the same value but as separate
        children of each of the first levels. GitLab issue #391
        """
        self._create_test_document_stub()

        # Create simple index template
        level_1 = self.test_index_template.index_template_nodes.create(
            expression='{{ document.uuid }}', link_documents=False,
            parent=self.test_index_template_root_node
        )

        self.test_index_template.index_template_nodes.create(
            expression='{{ document.label }}', link_documents=True,
            parent=level_1
        )

        IndexTemplate.objects.rebuild()

        # Typecast to sets to make sorting irrelevant in the comparison.
        self.assertEqual(
            set(IndexInstanceNode.objects.values_list('value', flat=True)),
            set(
                [
                    '', str(self.test_documents[1].uuid), self.test_documents[1].label,
                    str(self.test_documents[0].uuid), self.test_documents[0].label
                ]
            )
        )

    def test_document_description_index(self):
        self._create_test_index_template_node(
            expression=TEST_INDEX_TEMPLATE_DOCUMENT_DESCRIPTION_EXPRESSION
        )

        self.test_document.description = TEST_DOCUMENT_DESCRIPTION
        self.test_document.save()

        self.test_index_template.rebuild()

        self.assertEqual(
            IndexInstanceNode.objects.last().value,
            self.test_document.description
        )
        self.test_document.description = TEST_DOCUMENT_DESCRIPTION_EDITED
        self.test_document.save()

        self.assertEqual(
            IndexInstanceNode.objects.last().value,
            self.test_document.description
        )

    def test_document_label_index(self):
        self._create_test_index_template_node(
            expression=TEST_INDEX_TEMPLATE_DOCUMENT_LABEL_EXPRESSION
        )

        self.test_index_template.rebuild()

        self.assertEqual(
            IndexInstanceNode.objects.last().value, self.test_document.label
        )
        self.test_document.label = TEST_DOCUMENT_LABEL_EDITED
        self.test_document.save()

        self.assertEqual(
            IndexInstanceNode.objects.last().value, self.test_document.label
        )

    def test_document_type_index(self):
        self._create_test_index_template_node(
            expression=TEST_INDEX_TEMPLATE_DOCUMENT_TYPE_EXPRESSION
        )

        self.test_index_template.rebuild()

        self.assertEqual(
            IndexInstanceNode.objects.last().value,
            self.test_document_types[0].label
        )

        self._create_test_document_type()
        self.test_index_template.document_types.add(self.test_document_type)

        self.test_document.document_type_change(
            document_type=self.test_document_type
        )

        self.assertEqual(
            IndexInstanceNode.objects.last().value,
            self.test_document_types[1].label
        )

    def test_metadata_indexing(self):
        metadata_type = MetadataType.objects.create(
            name=TEST_METADATA_TYPE_NAME, label=TEST_METADATA_TYPE_LABEL
        )
        DocumentTypeMetadataType.objects.create(
            document_type=self.test_document_type,
            metadata_type=metadata_type
        )

        self._create_test_index_template_node(
            expression=TEST_INDEX_TEMPLATE_METADATA_EXPRESSION
        )

        # Add document metadata value to trigger index node instance
        # creation.
        self.test_document.metadata.create(
            metadata_type=metadata_type, value='0001'
        )
        self.assertEqual(
            list(
                IndexInstanceNode.objects.values_list('value', flat=True)
            ), ['', '0001']
        )

        # Check that document is in instance node.
        instance_node = IndexInstanceNode.objects.get(value='0001')
        self.assertQuerysetEqual(
            instance_node.documents.all(), [repr(self.test_document)]
        )

        # Change document metadata value to trigger index node instance
        # update.
        document_metadata = self.test_document.metadata.get(
            metadata_type=metadata_type
        )
        document_metadata.value = '0002'
        document_metadata.save()
        self.assertEqual(
            list(
                IndexInstanceNode.objects.values_list('value', flat=True)
            ), ['', '0002']
        )

        # Check that document is in new instance node.
        instance_node = IndexInstanceNode.objects.get(value='0002')
        self.assertQuerysetEqual(
            instance_node.documents.all(), [repr(self.test_document)]
        )

        # Check node instance is destoyed when no metadata is available.
        self.test_document.metadata.get(metadata_type=metadata_type).delete()
        self.assertEqual(
            list(
                IndexInstanceNode.objects.values_list('value', flat=True)
            ), ['']
        )

        # Add document metadata value again to trigger index node instance
        # creation.
        self.test_document.metadata.create(
            metadata_type=metadata_type, value='0003'
        )
        self.assertEqual(
            list(
                IndexInstanceNode.objects.values_list('value', flat=True)
            ), ['', '0003']
        )

        # Check node instance is destroyed when no documents are contained.
        self.test_document.delete()

        # Document is in trash, index structure should remain unchanged.
        self.assertEqual(
            list(
                IndexInstanceNode.objects.values_list('value', flat=True)
            ), ['', '0003']
        )

        # Document deleted, index structure should update.
        self.test_document.delete()
        self.assertEqual(
            list(
                IndexInstanceNode.objects.values_list('value', flat=True)
            ), ['']
        )

    def test_method_get_absolute_url(self):
        test_index_instance = IndexInstance.objects.first()
        self.assertTrue(test_index_instance.get_absolute_url())

    def test_multi_level_template_with_no_result_parent(self):
        """
        On a two level template if the first level doesn't return a result
        the indexing should stop. GitLab issue #391.
        """
        level_1 = self.test_index_template.index_template_nodes.create(
            expression='', link_documents=True,
            parent=self.test_index_template_root_node,
        )

        self.test_index_template.index_template_nodes.create(
            expression='{{ document.label }}', link_documents=True,
            parent=level_1
        )

        IndexTemplate.objects.rebuild()

    def test_rebuild_all_indexes(self):
        # Add metadata type and connect to document type.
        metadata_type = MetadataType.objects.create(
            name='test', label='test'
        )
        DocumentTypeMetadataType.objects.create(
            document_type=self.test_document_type,
            metadata_type=metadata_type
        )

        # Add document metadata value.
        self.test_document.metadata.create(
            metadata_type=metadata_type, value='0001'
        )

        self._create_test_index_template_node(
            expression='{{ document.metadata_value_of.test }}'
        )

        self.assertEqual(
            list(
                IndexTemplateNode.objects.values_list('expression', flat=True)
            ), ['', '{{ document.metadata_value_of.test }}']
        )

        # There should be only a root index instances nodes.
        self.assertEqual(IndexInstanceNode.objects.count(), 1)
        self.assertEqual(IndexInstanceNode.objects.first().parent, None)

        # Rebuild all indexes.
        IndexTemplate.objects.rebuild()

        # Check that document is in instance node.
        instance_node = IndexInstanceNode.objects.get(value='0001')
        self.assertQuerysetEqual(
            instance_node.documents.all(), [repr(self.test_document)]
        )


class IndexIntegrityTestCase(
    IndexTemplateTestMixin, GenericTransactionDocumentTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_unique_value_per_level(self):
        self._create_test_index_template_node(
            expression=TEST_INDEX_TEMPLATE_DOCUMENT_LABEL_EXPRESSION
        )

        self.test_index_template.rebuild()

        index_instance_node = IndexInstanceNode.objects.last()

        with self.assertRaises(expected_exception=IntegrityError):
            IndexInstanceNode.objects.create(
                parent=index_instance_node.parent,
                index_template_node=index_instance_node.index_template_node,
                value=index_instance_node.value
            )

        # Reset the failed database write to allow the database manager
        # to flush the database during the test tear down.
        IndexInstanceNode.objects.create(
            parent=index_instance_node.parent,
            index_template_node=index_instance_node.index_template_node,
            value='{}_{}'.format(
                index_instance_node.value, index_instance_node.pk
            )
        )



import os
from urllib.parse import quote

from django.conf import settings
from django.http.response import FileResponse as DjangoFileResponse

from mayan.apps.mimetype.api import get_mimetype


class FileResponse(DjangoFileResponse):
    """
    Subclass of Django's 2.2 FileResponse
    Modified to allows downloading non file like content as attachment
    A streaming HTTP response class optimized for files.
    """
    def _set_as_attachment(self, filename):
        if self.as_attachment:
            filename = self.filename or os.path.basename(filename)
            if filename:
                try:
                    filename.encode('ascii')
                    file_expr = 'filename="{}"'.format(filename)
                except UnicodeEncodeError:
                    file_expr = "filename*=utf-8''{}".format(quote(filename))
                self['Content-Disposition'] = 'attachment; {}'.format(file_expr)

    def _set_streaming_content(self, value):
        if not hasattr(value, 'read'):
            self.file_to_stream = None
            result = super()._set_streaming_content(value)
            self._set_as_attachment(filename=self.filename)

            return result

        self.file_to_stream = filelike = value
        if hasattr(filelike, 'close'):
            self._closable_objects.append(filelike)
        value = iter(lambda: filelike.read(self.block_size), b'')
        self.set_headers(filelike)
        super()._set_streaming_content(value)

    def set_headers(self, filelike):
        """
        Set some common response headers (Content-Length, Content-Type, and
        Content-Disposition) based on the `filelike` response content.
        """
        encoding_map = {
            'bzip2': 'application/x-bzip',
            'gzip': 'application/gzip',
            'xz': 'application/x-xz',
        }
        filename = getattr(filelike, 'name', None)
        filename = filename if (isinstance(filename, str) and filename) else self.filename
        if os.path.isabs(filename):
            self['Content-Length'] = os.path.getsize(filelike.name)
        elif hasattr(filelike, 'getbuffer'):
            self['Content-Length'] = filelike.getbuffer().nbytes

        if self.get('Content-Type', '').startswith(settings.DEFAULT_CONTENT_TYPE):
            if self.file_to_stream:
                content_type, encoding = get_mimetype(
                    file_object=self.file_to_stream, mimetype_only=True
                )
                # Encoding isn't set to prevent browsers from automatically
                # uncompressing files.
                content_type = encoding_map.get(encoding, content_type)
                self['Content-Type'] = content_type or 'application/octet-stream'
            else:
                self['Content-Type'] = 'application/octet-stream'

        self._set_as_attachment(filename=filename)



from django_celery_beat.models import PeriodicTask

from mayan.apps.documents.tests.base import GenericDocumentTestCase

from .mixins.base_mixins import PeriodicSourceBackendTestMixin


class PeriodicSourceBackendTestCase(
    PeriodicSourceBackendTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    def test_periodic_source_delete(self):
        periodic_task_count = PeriodicTask.objects.count()

        self.test_source.delete()

        self.assertEqual(PeriodicTask.objects.count(), periodic_task_count - 1)

    def test_periodic_source_save(self):
        periodic_task_count = PeriodicTask.objects.count()

        self.test_source.save()

        self.assertEqual(PeriodicTask.objects.count(), periodic_task_count)



from mayan.apps.appearance.classes import Icon

# Document version

icon_document_version_ocr_content_delete_single = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='times'
)
icon_document_version_ocr_content_delete_multiple = icon_document_version_ocr_content_delete_single
icon_document_version_ocr_content_view = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_version_ocr_submit_single = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='arrow-right'
)
icon_document_version_ocr_submit_multiple = icon_document_version_ocr_submit_single

# Document version page

icon_document_version_page_ocr_content_detail_view = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_version_page_ocr_content_edit_view = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_version_ocr_download = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='arrow-down'
)
icon_document_version_ocr_errors_list = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='exclamation'
)

# Document type

icon_document_type_ocr_settings = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_type_submit = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='arrow-right'
)

icon_entry_list = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='exclamation'
)



from rest_framework.authtoken.models import Token

from mayan.apps.testing.tests.mixins import TestViewTestCaseMixin


class APIUserTestCaseMixin:
    def setUp(self):
        super().setUp()
        self._test_case_user_token = self.get_test_user_token(
            user=self._test_case_user
        )

    def get_test_user_token(self, user):
        token, created = Token.objects.get_or_create(
            user=user
        )

        return token


class RESTAPIViewTestMixin:
    def _request_test_browser_api_view(self):
        return self.get(query={'format': 'api'}, viewname='rest_api:api_root')

    def _request_test_redoc_ui_view(self):
        return self.get(viewname='rest_api:schema-redoc')

    def _request_test_swagger_ui_view(self):
        return self.get(viewname='rest_api:schema-swagger-ui')

    def _request_test_swagger_no_ui_json_view(self):
        return self.get(
            kwargs={'format': '.json'}, viewname='rest_api:schema-json'
        )

    def _request_test_swagger_no_ui_yaml_view(self):
        return self.get(
            kwargs={'format': '.yaml'}, viewname='rest_api:schema-json'
        )


class TestAPIViewTestCaseMixin(TestViewTestCaseMixin):
    def _get_test_view_urlpatterns(self):
        from ..urls import api_version_urls
        return api_version_urls



from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..permissions import permission_cabinet_view

from .mixins import CabinetSearchTestMixin, CabinetTestMixin


class CabinetSearchTestCase(
    CabinetTestMixin, CabinetSearchTestMixin, GenericDocumentViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_cabinet(add_test_document=True)

    def test_cabinet_search_no_permission(self):
        queryset = self._perform_cabinet_search()
        self.assertFalse(self.test_cabinet in queryset)

    def test_cabinet_search_with_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )

        queryset = self._perform_cabinet_search()
        self.assertTrue(self.test_cabinet in queryset)



from django.core import mail

from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..events import event_email_sent
from ..permissions import (
    permission_send_document_file_attachment,
    permission_send_document_file_link, permission_user_mailer_use
)

from .literals import TEST_EMAIL_ADDRESS, TEST_EMAIL_FROM_ADDRESS
from .mixins import DocumentFileMailerViewTestMixin, MailerTestMixin


class MailDocumentFileViewTestCase(
    DocumentFileMailerViewTestMixin, MailerTestMixin,
    GenericDocumentViewTestCase
):
    def test_document_file_send_link_single_view_no_permission(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self._clear_events()

        response = self._request_test_document_file_send_link_single_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_single_view_with_document_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_link
        )

        self._clear_events()

        response = self._request_test_document_file_send_link_single_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_single_view_with_mailer_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self._clear_events()

        response = self._request_test_document_file_send_link_single_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_single_view_with_full_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_link
        )
        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self._clear_events()

        response = self._request_test_document_file_send_link_single_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(len(mail.outbox), mail_messages + 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_file)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_trashed_document_file_send_link_single_view_with_full_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_link
        )
        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_send_link_single_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_multiple_view_no_permission(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self._clear_events()

        response = self._request_test_document_file_send_link_multiple_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_multiple_view_with_document_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_link
        )

        self._clear_events()

        response = self._request_test_document_file_send_link_multiple_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_multiple_view_with_mailer_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self._clear_events()

        response = self._request_test_document_file_send_link_multiple_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_multiple_view_with_full_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_link
        )
        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self._clear_events()

        response = self._request_test_document_file_send_link_multiple_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(len(mail.outbox), mail_messages + 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_file)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_trashed_document_file_send_link_multiple_view_with_full_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_link
        )
        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_send_link_multiple_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_attachment_send_single_view_no_permission(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self._clear_events()

        response = self._request_test_document_file_attachment_send_single_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_attachment_send_single_view_with_document_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_attachment
        )

        self._clear_events()

        response = self._request_test_document_file_attachment_send_single_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_attachment_send_single_view_with_mailer_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self._clear_events()

        response = self._request_test_document_file_attachment_send_single_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_attachment_send_single_view_with_full_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_attachment
        )
        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self._clear_events()

        response = self._request_test_document_file_attachment_send_single_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(len(mail.outbox), mail_messages + 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])
        self.assertEqual(len(mail.outbox[0].attachments), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_file)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_trashed_document_file_attachment_send_single_view_with_full_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_attachment
        )
        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_attachment_send_single_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_attachment_send_multiple_view_no_permission(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self._clear_events()

        response = self._request_test_document_file_attachment_send_multiple_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_attachment_send_multiple_view_with_document_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_attachment
        )

        self._clear_events()

        response = self._request_test_document_file_attachment_send_multiple_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_attachment_send_multiple_view_with_mailer_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self._clear_events()

        response = self._request_test_document_file_attachment_send_multiple_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_attachment_send_multiple_view_with_full_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_attachment
        )
        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self._clear_events()

        response = self._request_test_document_file_attachment_send_multiple_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(len(mail.outbox), mail_messages + 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])
        self.assertEqual(len(mail.outbox[0].attachments), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_file)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_trashed_document_file_attachment_send_multiple_view_with_full_access(self):
        self._create_test_user_mailer()

        mail_messages = len(mail.outbox)

        self.grant_access(
            obj=self.test_document,
            permission=permission_send_document_file_attachment
        )
        self.grant_access(
            obj=self.test_user_mailer, permission=permission_user_mailer_use
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_file_attachment_send_multiple_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(len(mail.outbox), mail_messages)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.conf.urls import url

from .api_views import (
    APIEventListView, APIEventTypeListView, APIEventTypeNamespaceDetailView,
    APIEventTypeNamespaceEventTypeListView, APIEventTypeNamespaceListView,
    APINotificationListView, APIObjectEventListView
)
from .views.clear_views import (
    EventListClearView, ObjectEventClearView, VerbEventClearView,
    CurrentUserEventClearView,
)
from .views.event_views import (
    CurrentUserEventListView, EventListView, ObjectEventListView,
    VerbEventListView
)
from .views.export_views import (
    CurrentUserEventExportView, EventListExportView, ObjectEventExportView,
    VerbEventExportView
)
from .views.notification_views import (
    NotificationListView, NotificationMarkRead, NotificationMarkReadAll
)
from .views.subscription_views import (
    EventTypeSubscriptionListView, ObjectEventTypeSubscriptionListView
)


urlpatterns_events = [
    url(regex=r'^events/$', name='events_list', view=EventListView.as_view()),
    url(
        regex=r'^object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/events/$',
        name='events_for_object', view=ObjectEventListView.as_view()
    ),
    url(
        regex=r'^user/events/$', name='current_user_events',
        view=CurrentUserEventListView.as_view()
    ),
    url(
        regex=r'^verbs/(?P<verb>[\w\-\.]+)/$', name='events_by_verb',
        view=VerbEventListView.as_view(),
    )
]

urlpatterns_events_clear = [
    url(
        regex=r'^events/clear/$', name='events_list_clear',
        view=EventListClearView.as_view()
    ),
    url(
        regex=r'^object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/events/clear/$',
        name='events_for_object_clear', view=ObjectEventClearView.as_view()
    ),
    url(
        regex=r'^verbs/(?P<verb>[\w\-\.]+)/clear/$', name='events_by_verb_clear',
        view=VerbEventClearView.as_view(),
    ),
    url(
        regex=r'^user/events/clear/$', name='current_user_events_clear',
        view=CurrentUserEventClearView.as_view()
    )
]

urlpatterns_events_export = [
    url(
        regex=r'^events/export/$', name='events_list_export',
        view=EventListExportView.as_view()
    ),
    url(
        regex=r'^object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/events/export/$',
        name='events_for_object_export', view=ObjectEventExportView.as_view()
    ),
    url(
        regex=r'^verbs/(?P<verb>[\w\-\.]+)/export/$', name='events_by_verb_export',
        view=VerbEventExportView.as_view(),
    ),
    url(
        regex=r'^user/events/export/$', name='current_user_events_export',
        view=CurrentUserEventExportView.as_view()
    )
]

urlpatterns_notification = [
    url(
        regex=r'^user/notifications/$', name='user_notifications_list',
        view=NotificationListView.as_view()
    ),
    url(
        regex=r'^user/notifications/(?P<notification_id>\d+)/mark_read/$',
        name='notification_mark_read', view=NotificationMarkRead.as_view()
    ),
    url(
        regex=r'^user/notifications/all/mark_read/$',
        name='notification_mark_read_all',
        view=NotificationMarkReadAll.as_view()
    ),
]

urlpatterns_subscriptions = [
    url(
        regex=r'^user/event_types/subscriptions/$',
        name='event_types_user_subscriptions_list',
        view=EventTypeSubscriptionListView.as_view()
    ),
    url(
        regex=r'^user/object/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/subscriptions/$',
        name='object_event_types_user_subscriptions_list',
        view=ObjectEventTypeSubscriptionListView.as_view()
    ),
]

urlpatterns = []
urlpatterns.extend(urlpatterns_events)
urlpatterns.extend(urlpatterns_events_clear)
urlpatterns.extend(urlpatterns_events_export)
urlpatterns.extend(urlpatterns_notification)
urlpatterns.extend(urlpatterns_subscriptions)

api_urls = [
    url(
        regex=r'^event_type_namespaces/(?P<name>[-\w]+)/$',
        name='event-type-namespace-detail',
        view=APIEventTypeNamespaceDetailView.as_view()
    ),
    url(
        regex=r'^event_type_namespaces/(?P<name>[-\w]+)/event_types/$',
        name='event-type-namespace-event-type-list',
        view=APIEventTypeNamespaceEventTypeListView.as_view()
    ),
    url(
        regex=r'^event_type_namespaces/$',
        name='event-type-namespace-list',
        view=APIEventTypeNamespaceListView.as_view()
    ),
    url(
        regex=r'^event_types/$', name='event-type-list',
        view=APIEventTypeListView.as_view()
    ),
    url(
        regex=r'^events/$', name='event-list', view=APIEventListView.as_view()
    ),
    url(
        regex=r'^notifications/$', name='notification-list',
        view=APINotificationListView.as_view()
    ),
    url(
        regex=r'^objects/(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)/(?P<object_id>\d+)/events/$',
        name='object-event-list', view=APIObjectEventListView.as_view()
    ),
]



from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured

from rest_framework import status
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response
from rest_framework.settings import api_settings

from mayan.apps.databases.utils import check_queryset
from mayan.apps.views.mixins import ExternalObjectBaseMixin


class AsymmetricSerializerAPIViewMixin:
    _write_methods = ('PATCH', 'POST', 'PUT')
    read_serializer_class = None
    write_serializer_class = None

    def get_read_serializer_class(self):
        if not self.read_serializer_class:
            raise ImproperlyConfigured(
                'View {} must specify a read_serializer_class.'.format(
                    self.__class__.__name__
                )
            )
        else:
            return self.read_serializer_class

    def get_serializer_class(self):
        if getattr(self, 'serializer_class', None):
            raise ImproperlyConfigured(
                'View {} can not use AsymmetricSerializerAPIViewMixin while '
                'also specifying a serializer_class.'.format(
                    self.__class__.__name__
                )
            )

        if self.request.method in self._write_methods:
            return self.get_write_serializer_class()
        else:
            return self.get_read_serializer_class()

    def get_write_serializer_class(self):
        if not self.write_serializer_class:
            raise ImproperlyConfigured(
                'View {} must specify a write_serializer_class.'.format(
                    self.__class__.__name__
                )
            )
        else:
            return self.write_serializer_class


class CheckQuerysetAPIViewMixin:
    def get_queryset(self, *args, **kwargs):
        queryset = super().get_queryset(*args, **kwargs)
        return check_queryset(self=self, queryset=queryset)


class ContentTypeAPIViewMixin:
    """
    This mixin makes it easier for API views to retrieve a content type from
    the URL pattern.
    """
    content_type_url_kw_args = {
        'app_label': 'app_label',
        'model_name': 'model_name'
    }

    def get_content_type(self):
        return get_object_or_404(
            queryset=ContentType, app_label=self.kwargs[
                self.content_type_url_kw_args['app_label']
            ], model=self.kwargs[
                self.content_type_url_kw_args['model_name']
            ]
        )


class ExternalObjectAPIViewMixin(ExternalObjectBaseMixin):
    """
    Override get_external_object to use REST API get_object_or_404.
    """
    def initial(self, *args, **kwargs):
        result = super().initial(*args, **kwargs)
        # Ensure self.external_object is initialized to allow the browseable
        # API view to display when attempting to introspect the serializer
        # and the parent object is not found.
        self.external_object = None

        self.external_object = self.get_external_object()
        return result

    def get_serializer_extra_context(self):
        """
        Add the external object to the serializer context. Useful for the
        create action when there is no instance available.
        """
        result = {}
        if self.kwargs:
            result['external_object'] = self.external_object

        return result

    def get_external_object(self):
        return get_object_or_404(
            queryset=self.get_external_object_queryset_filtered(),
            **self.get_pk_url_kwargs()
        )

    def get_external_object_permission(self):
        return getattr(
            self, 'mayan_external_object_permissions', {}
        ).get(self.request.method, (None,))[0]


class ExternalContentTypeObjectAPIViewMixin(
    ContentTypeAPIViewMixin, ExternalObjectAPIViewMixin
):
    """
    Mixin to retrieve an external object by content type from the URL pattern.
    """
    external_object_pk_url_kwarg = 'object_id'

    def get_external_object_queryset(self):
        self.content_type = self.get_content_type()
        self.external_object_class = self.content_type.model_class()
        return super().get_external_object_queryset()


class InstanceExtraDataAPIViewMixin:
    def perform_create(self, serializer):
        if hasattr(self, 'get_instance_extra_data'):
            serializer.validated_data['_instance_extra_data'] = self.get_instance_extra_data()

        return super().perform_create(serializer=serializer)

    def perform_destroy(self, instance):
        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                setattr(instance, key, value)

        return super().perform_destroy(instance=instance)

    def perform_update(self, serializer):
        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                serializer.validated_data[key] = value

        return super().perform_update(serializer=serializer)


class SchemaInspectionAPIViewMixin:
    def get_queryset(self):
        if getattr(self, 'swagger_fake_view', False):
            return []

        return super().get_queryset()

    def get_serializer(self, *args, **kwargs):
        if getattr(self, 'swagger_fake_view', False):

            return None

        return super().get_serializer(*args, **kwargs)

    def get_serializer_context(self, *args, **kwargs):
        if getattr(self, 'swagger_fake_view', False):
            return {}

        return super().get_serializer_context(*args, **kwargs)


class SerializerActionAPIViewMixin:
    serializer_action_name = None

    def serializer_action(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_serializer_action(serializer=serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(
            serializer.data, status=status.HTTP_200_OK, headers=headers
        )

    def get_success_headers(self, data):
        try:
            return {'Location': str(data[api_settings.URL_FIELD_NAME])}
        except (TypeError, KeyError):
            return {}

    def perform_serializer_action(self, serializer):
        getattr(serializer, self.serializer_action_name)()

    def post(self, request, *args, **kwargs):
        return self.serializer_action(request=request, *args, **kwargs)


class SerializerExtraContextAPIViewMixin:
    def get_serializer_extra_context(self):
        return {}

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context.update(self.get_serializer_extra_context())
        return context


class SuccessHeadersAPIViewMixin:
    def get_success_headers(self, data):
        try:
            return {'Location': str(data[api_settings.URL_FIELD_NAME])}
        except (TypeError, KeyError):
            return {}



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0021_auto_20210307_0439'),
    ]

    operations = [
        migrations.CreateModel(
            name='IndexInstance',
            fields=[
            ],
            options={
                'verbose_name': 'Index instance',
                'verbose_name_plural': 'Index instances',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('document_indexing.indextemplate',),
        ),
    ]



from mayan.apps.events.classes import EventModelRegistry
from mayan.apps.testing.tests.base import BaseTestCase

from .mixins import DownloadFileTestMixin


class DownloadFileModelTestCase(DownloadFileTestMixin, BaseTestCase):
    def test_method_get_absolute_url_without_content_object(self):
        self._create_test_download_file()

        self.assertFalse(self.test_download_file.get_absolute_url())

    def test_method_get_absolute_url_with_content_object(self):
        self._create_test_object()
        self.TestModel.add_to_class(
            name='get_absolute_url', value=lambda self: 'test_value'
        )
        EventModelRegistry.register(model=self.TestModel)

        self._create_test_download_file(content_object=self.test_object)

        self.assertTrue(self.test_download_file.get_absolute_url())



from django.core import mail

from mayan.apps.documents.literals import DOCUMENT_VERSION_EXPORT_MIMETYPE
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.testing.tests.base import BaseTestCase

from ..events import event_email_sent
from ..literals import MODEL_SEND_FUNCTION_DOTTED_PATH

from .literals import (
    TEST_EMAIL_BODY_HTML, TEST_EMAIL_ADDRESS, TEST_EMAIL_FROM_ADDRESS,
    TEST_RECIPIENTS_MULTIPLE_COMMA, TEST_RECIPIENTS_MULTIPLE_COMMA_RESULT,
    TEST_RECIPIENTS_MULTIPLE_SEMICOLON,
    TEST_RECIPIENTS_MULTIPLE_SEMICOLON_RESULT, TEST_RECIPIENTS_MULTIPLE_MIXED,
    TEST_RECIPIENTS_MULTIPLE_MIXED_RESULT,
)
from .mixins import MailerTestMixin


class MailingModelTestCase(MailerTestMixin, BaseTestCase):
    def test_method_get_absolute_url(self):
        self._create_test_user_mailer()

        self.assertTrue(self.test_user_mailer.get_absolute_url())

    def test_send_simple(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send(to=TEST_EMAIL_ADDRESS)

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_simple_with_html(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send(
            to=TEST_EMAIL_ADDRESS, body=TEST_EMAIL_BODY_HTML
        )

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])
        self.assertEqual(
            mail.outbox[0].alternatives[0][0], TEST_EMAIL_BODY_HTML
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_multiple_recipients_comma(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send(to=TEST_RECIPIENTS_MULTIPLE_COMMA)

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(
            mail.outbox[0].to, TEST_RECIPIENTS_MULTIPLE_COMMA_RESULT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_multiple_recipients_semicolon(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send(to=TEST_RECIPIENTS_MULTIPLE_SEMICOLON)

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(
            mail.outbox[0].to, TEST_RECIPIENTS_MULTIPLE_SEMICOLON_RESULT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_multiple_recipient_mixed(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send(to=TEST_RECIPIENTS_MULTIPLE_MIXED)

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(
            mail.outbox[0].to, TEST_RECIPIENTS_MULTIPLE_MIXED_RESULT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_to_cc(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send(
            cc=TEST_EMAIL_ADDRESS, to=TEST_EMAIL_ADDRESS
        )

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])
        self.assertEqual(mail.outbox[0].cc, [TEST_EMAIL_ADDRESS])

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_to_bcc(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send(
            bcc=TEST_EMAIL_ADDRESS, to=TEST_EMAIL_ADDRESS
        )

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])
        self.assertEqual(mail.outbox[0].bcc, [TEST_EMAIL_ADDRESS])

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_with_reply_to(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send(
            reply_to=TEST_EMAIL_ADDRESS, to=TEST_EMAIL_ADDRESS
        )

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])
        self.assertEqual(mail.outbox[0].reply_to, [TEST_EMAIL_ADDRESS])

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)


class DocumentMailingTestCase(MailerTestMixin, GenericDocumentTestCase):
    def test_send_link(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send_object(
            obj=self.test_document, to=TEST_EMAIL_ADDRESS
        )

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)


class DocumentFileMailingTestCase(MailerTestMixin, GenericDocumentTestCase):
    def test_send_document_file_attachment(self):
        self._create_test_user_mailer()

        self._clear_events()

        kwargs = {
            'as_attachment': True, 'obj': self.test_document_file,
            'to': TEST_EMAIL_ADDRESS
        }
        kwargs.update(
            MODEL_SEND_FUNCTION_DOTTED_PATH.get(
                self.test_document_file._meta.model, {}
            )
        )

        self.test_user_mailer.send_object(**kwargs)

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])
        self.assertEqual(
            mail.outbox[0].attachments[0][0],
            self.test_document_file.filename
        )
        self.assertEqual(
            mail.outbox[0].attachments[0][2],
            self.test_document_file.mimetype
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_file)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_document_file_link(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send_object(
            obj=self.test_document_file, to=TEST_EMAIL_ADDRESS
        )

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])
        self.assertEqual(len(mail.outbox[0].attachments), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_file)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)


class DocumentVersionMailingTestCase(
    MailerTestMixin, GenericDocumentTestCase
):
    def test_send_document_version_attachment(self):
        self._create_test_user_mailer()

        self._clear_events()

        kwargs = {
            'as_attachment': True, 'obj': self.test_document_version,
            'to': TEST_EMAIL_ADDRESS
        }
        kwargs.update(
            MODEL_SEND_FUNCTION_DOTTED_PATH.get(
                self.test_document_version._meta.model, {}
            )
        )

        self.test_user_mailer.send_object(**kwargs)

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])
        self.assertEqual(
            mail.outbox[0].attachments[0][0],
            str(self.test_document_version)
        )
        self.assertEqual(
            mail.outbox[0].attachments[0][2],
            DOCUMENT_VERSION_EXPORT_MIMETYPE
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_version)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_send_document_version_link(self):
        self._create_test_user_mailer()

        self._clear_events()

        self.test_user_mailer.send_object(
            obj=self.test_document_version,
            to=TEST_EMAIL_ADDRESS
        )

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS)
        self.assertEqual(mail.outbox[0].to, [TEST_EMAIL_ADDRESS])
        self.assertEqual(len(mail.outbox[0].attachments), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_version)
        self.assertEqual(events[0].actor, self.test_user_mailer)
        self.assertEqual(events[0].target, self.test_user_mailer)
        self.assertEqual(events[0].verb, event_email_sent.id)



import sentry_sdk
from sentry_sdk.integrations.django import DjangoIntegration

from django.conf.urls import url

import mayan

from .classes import ClientBackend


class ClientBackendSentry(ClientBackend):
    _url_namespace = 'sentry'

    def get_url_patterns(self):
        def view_trigger_error(request):
            1 / 0

        return [
            url(
                regex=r'^debug/$', name='sentry_debug',
                view=view_trigger_error
            )
        ]

    def launch(self):
        self.setup_arguments()

        sentry_sdk.init(
            dsn=self.dsn,
            integrations=(DjangoIntegration(),),
            traces_sample_rate=self.traces_sample_rate,
            send_default_pii=self.send_default_pii,
            release=self.release
        )

    def setup_arguments(self):
        self.dsn = self.kwargs['dsn']

        # Set traces_sample_rate to 1.0 to capture 100%
        # of transactions for performance monitoring.
        # We recommend adjusting this value in production,
        self.traces_sample_rate = self.kwargs.get('traces_sample_rate', 1.0)

        # If you wish to associate users to errors (assuming you are using
        # django.contrib.auth) you may enable sending PII data.
        self.send_default_pii = self.kwargs.get('send_default_pii', True)

        # By default the SDK will try to use the SENTRY_RELEASE
        # environment variable, or infer a git commit
        # SHA as release, however you may want to set
        # something more human-readable.
        self.release = mayan.__build_string__



from django.core.exceptions import ImproperlyConfigured
from django.template.loader import render_to_string

from mayan.apps.templating.classes import Template


class SourceColumnWidget:
    template_name = None
    template_string = None

    def __init__(self, column, request):
        self.column = column
        self.request = request

    def get_extra_context(self):
        return {}

    def get_template_name(self):
        return self.template_name

    def get_template_string(self):
        return self.template_string

    def render(self, value=None):
        template_name = self.get_template_name()
        template_string = self.get_template_string()

        self.value = value
        context = {
            'column': self.column, 'value': value
        }
        context.update(self.get_extra_context())

        if template_name:
            return render_to_string(
                template_name=self.template_name, context=context
            )
        elif template_string:
            return Template(template_string=template_string).render(context=context)
        else:
            raise ImproperlyConfigured(
                'SourceColumnWidget `{}` must provide either '
                '`template_name`, `template_string`, `get_template_name()`, '
                'or `get_template_string()`.'.format(
                    self.__class__.__name__
                )
            )


class SourceColumnLinkWidget(SourceColumnWidget):
    template_name = 'navigation/source_column_link_widget.html'



from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..links import link_document_create_multiple


class SourcesLinksTestCase(GenericDocumentViewTestCase):
    def _get_document_create_link(self):
        self.add_test_view(test_object=self.test_document)
        context = self.get_test_view()
        context['user'] = self._test_case_user
        return link_document_create_multiple.resolve(context=context)

    def test_document_create_link_no_permission(self):
        resolved_link = self._get_document_create_link()
        self.assertEqual(resolved_link, None)

    def test_document_create_link_with_access(self):
        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )
        resolved_link = self._get_document_create_link()
        self.assertNotEqual(resolved_link, None)



from django.db import migrations, models
from django.utils.text import slugify

from mayan.apps.common.validators import validate_internal_name


def operation_generate_internal_name(apps, schema_editor):
    Workflow = apps.get_model(
        app_label='document_states', model_name='Workflow'
    )
    internal_names = []

    for workflow in Workflow.objects.using(alias=schema_editor.connection.alias).all():
        # Slugify and replace dashes (not allowed) by underscores
        workflow.internal_name = slugify(workflow.label).replace('-', '_')
        if workflow.internal_name in internal_names:
            # Add a suffix in case two conversions yield the same
            # result.
            workflow.internal_name = '{}_'.format(
                workflow.internal_name
            )

        internal_names.append(workflow.internal_name)
        workflow.save()


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0003_auto_20170325_0447'),
    ]

    operations = [
        # Add the internal name field but make it non unique
        # https://docs.djangoproject.com/en/1.10/howto/
        # writing-migrations/#migrations-that-add-unique-fields
        migrations.AddField(
            model_name='workflow',
            name='internal_name',
            field=models.CharField(
                db_index=False, default=' ',
                help_text='This value will be used by other apps to reference '
                'this workflow. Can only contain letters, numbers, and '
                'underscores.', max_length=255, unique=False, validators=[
                    validate_internal_name
                ], verbose_name='Internal name'
            ),
        ),

        # Generate the slugs based on the labels
        migrations.RunPython(
            code=operation_generate_internal_name,
            reverse_code=migrations.RunPython.noop
        ),

        # Make the internal name field unique
        # Add the internal name field but make it non unique
        # https://docs.djangoproject.com/en/1.10/howto/
        # writing-migrations/#migrations-that-add-unique-fields
        migrations.AlterField(
            model_name='workflow',
            name='internal_name',
            field=models.CharField(
                db_index=True,
                help_text='This value will be used by other apps to reference '
                'this workflow. Can only contain letters, numbers, and '
                'underscores.', max_length=255, unique=True, validators=[
                    validate_internal_name
                ], verbose_name='Internal name'
            ),
        ),
    ]



import json
import os

from django import forms
from django.conf import settings
from django.contrib.admin.utils import label_for_field
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db import models
from django.db.models import Model
from django.db.models.query import QuerySet
from django.forms.models import ModelFormMetaclass
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.common.utils import introspect_attribute, resolve_attribute

from .widgets import DisableableSelectWidget, PlainWidget, TextAreaDiv


class ChoiceForm(forms.Form):
    """
    Form to be used in side by side templates used to add or remove
    items from a many to many field.
    """
    search = forms.CharField(
        label=_('Search'), required=False, widget=forms.widgets.TextInput(
            attrs={
                'autocomplete': 'off',
                'class': 'views-select-search',
                'placeholder': 'Filter list'
            }
        )
    )
    selection = forms.MultipleChoiceField(
        required=False, widget=DisableableSelectWidget()
    )

    def __init__(self, *args, **kwargs):
        choices = kwargs.pop('choices', [])
        label = kwargs.pop('label', _('Selection'))
        help_text = kwargs.pop('help_text', None)
        disabled_choices = kwargs.pop('disabled_choices', ())
        super().__init__(*args, **kwargs)
        self.fields['selection'].choices = choices
        self.fields['selection'].label = label
        self.fields['selection'].help_text = help_text
        self.fields['selection'].widget.disabled_choices = disabled_choices
        self.fields['selection'].widget.attrs.update(
            {
                'class': 'full-height input-hotkey-double-click',
                'data-height-difference': '495'
            }
        )


class FormOptions:
    def __init__(self, form, kwargs, options=None):
        """
        Option definitions will be iterated. The option value will be
        determined in the following order: as passed via keyword
        arguments during form intialization, as form get_... method or
        finally as static Meta options. This is to allow a form with
        Meta options or method to be overridden at initialization
        and increase the usability of a single class.
        """
        for name, default_value in self.option_definitions.items():
            try:
                # Check for a runtime value via kwargs
                value = kwargs.pop(name)
            except KeyError:
                try:
                    # Check if there is a get_... method
                    value = getattr(self, 'get_{}'.format(name))()
                except AttributeError:
                    try:
                        # Check the meta class options
                        value = getattr(options, name)
                    except AttributeError:
                        value = default_value

            setattr(self, name, value)


class DetailFormOption(FormOptions):
    # Dictionary list of option names and default values.
    option_definitions = {'extra_fields': []}


class DetailForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        self.opts = DetailFormOption(
            form=self, kwargs=kwargs, options=getattr(self, 'Meta', None)
        )
        super().__init__(*args, **kwargs)

        for extra_field in self.opts.extra_fields:
            obj = extra_field.get('object', self.instance)
            field = extra_field['field']

            result = resolve_attribute(
                attribute=field, obj=obj
            )

            label = extra_field.get('label', None)

            # If label is not specified try to get it from the object itself.
            if not label:
                attribute_name, obj = introspect_attribute(
                    attribute_name=field, obj=obj
                )

                if not obj:
                    label = _('None')
                else:
                    try:
                        label = getattr(
                            getattr(obj, attribute_name), 'short_description'
                        )
                    except AttributeError:
                        label = label_for_field(
                            name=attribute_name, model=obj
                        )

            help_text = extra_field.get('help_text', None)
            # If help_text is not specified try to get it from the object
            # itself.
            if not help_text:
                if obj:
                    try:
                        field_object = obj._meta.get_field(field_name=field)
                    except FieldDoesNotExist:
                        field_object = field

                    help_text = getattr(
                        field_object, 'help_text', None
                    )

            if isinstance(result, models.query.QuerySet):
                self.fields[field] = forms.ModelMultipleChoiceField(
                    queryset=result, label=label
                )
            else:
                self.fields[field] = forms.CharField(
                    initial=resolve_attribute(
                        obj=obj,
                        attribute=field
                    ), label=label, help_text=help_text,
                    widget=extra_field.get('widget', PlainWidget)
                )

        for field_name, field in self.fields.items():
            self.fields[field_name].widget.attrs.update(
                {'readonly': 'readonly'}
            )


class DynamicFormMixin:
    def __init__(self, *args, **kwargs):
        self.schema = kwargs.pop('schema')
        super().__init__(*args, **kwargs)

        widgets = self.schema.get('widgets', {})
        field_order = self.schema.get(
            'field_order', self.schema['fields'].keys()
        )

        for field_name in field_order:
            field_data = self.schema['fields'][field_name]
            field_class = import_string(dotted_path=field_data['class'])
            kwargs = {
                'label': field_data['label'],
                'required': field_data.get('required', True),
                'initial': field_data.get('default', None),
                'help_text': field_data.get('help_text'),
            }
            if widgets and field_name in widgets:
                widget = widgets[field_name]
                kwargs['widget'] = import_string(
                    dotted_path=widget['class']
                )(**widget.get('kwargs', {}))

            kwargs.update(field_data.get('kwargs', {}))
            self.fields[field_name] = field_class(**kwargs)

    @property
    def media(self):
        """
        Append the media of the dynamic fields to the normal fields' media.
        """
        media = super().media
        media = media + forms.Media(**self.schema.get('media', {}))
        return media


class DynamicForm(DynamicFormMixin, forms.Form):
    """Normal dynamic form."""


class DynamicModelForm(DynamicFormMixin, forms.ModelForm):
    """Dynamic model form."""


class DynamicFormMetaclass(ModelFormMetaclass):
    def __new__(mcs, name, bases, attrs):
        new_class = super(DynamicFormMetaclass, mcs).__new__(
            mcs=mcs, name=name, bases=bases, attrs=attrs
        )

        if new_class._meta.fields:
            new_class._meta.fields += ('backend_data',)
            widgets = getattr(new_class._meta, 'widgets', {}) or {}
            widgets['backend_data'] = forms.widgets.HiddenInput
            new_class._meta.widgets = widgets

        return new_class


class BackendDynamicForm(DynamicModelForm, metaclass=DynamicFormMetaclass):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        backend_data = self.instance.get_backend_data()

        if backend_data:
            for field_name in self.instance.get_backend().get_fields():
                self.fields[field_name].initial = backend_data.get(
                    field_name, None
                )

    def clean(self):
        data = super().clean()

        # Consolidate the dynamic fields into a single JSON field called
        # 'backend_data'.
        backend_data = {}

        for field_name, field_data in self.schema['fields'].items():
            backend_data[field_name] = data.pop(
                field_name, field_data.get('default', None)
            )
            if isinstance(backend_data[field_name], QuerySet):
                # Flatten the queryset to a list of ids.
                backend_data[field_name] = list(
                    backend_data[field_name].values_list('id', flat=True)
                )
            elif isinstance(backend_data[field_name], Model):
                # Store only the ID of a model instance.
                backend_data[field_name] = backend_data[field_name].pk

        data['backend_data'] = json.dumps(obj=backend_data)

        return data


class FileDisplayForm(forms.Form):
    DIRECTORY = None
    FILENAME = None

    text = forms.CharField(
        label='',
        widget=TextAreaDiv(
            attrs={
                'class': 'full-height scrollable',
                'data-height-difference': 270
            }
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.DIRECTORY or self.FILENAME:
            file_path = os.path.join(
                settings.BASE_DIR, os.sep.join(self.DIRECTORY), self.FILENAME
            )
            with open(file=file_path) as file_object:
                self.fields['text'].initial = file_object.read()


class FilteredSelectionFormOptions(FormOptions):
    # Dictionary list of option names and default values.
    option_definitions = {
        'allow_multiple': False,
        'field_name': None,
        'help_text': None,
        'label': None,
        'model': None,
        'permission': None,
        'queryset': None,
        'required': True,
        'user': None,
        'widget_class': None,
        'widget_attributes': {'size': '10'},
    }


class FilteredSelectionForm(forms.Form):
    """
    Form to select the from a list of choice filtered by access. Can be
    configure to allow single or multiple selection.
    """
    def __init__(self, *args, **kwargs):
        opts = FilteredSelectionFormOptions(
            form=self, kwargs=kwargs, options=getattr(self, 'Meta', None)
        )

        if opts.queryset is None:
            if not opts.model:
                raise ImproperlyConfigured(
                    '{} requires a queryset or a model to be specified as '
                    'a meta option or passed during initialization.'.format(
                        self.__class__.__name__
                    )
                )

            queryset = opts.model.objects.all()
        else:
            queryset = opts.queryset

        if opts.allow_multiple:
            extra_kwargs = {}
            field_class = forms.ModelMultipleChoiceField
            widget_class = forms.widgets.SelectMultiple
        else:
            extra_kwargs = {'empty_label': None}
            field_class = forms.ModelChoiceField
            widget_class = forms.widgets.Select

        if opts.widget_class:
            widget_class = opts.widget_class

        if opts.permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=opts.permission, queryset=queryset,
                user=opts.user
            )

        super().__init__(*args, **kwargs)

        self.fields[opts.field_name] = field_class(
            help_text=opts.help_text, label=opts.label,
            queryset=queryset, required=opts.required,
            widget=widget_class(attrs=opts.widget_attributes),
            **extra_kwargs
        )


class RelationshipForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self._event_actor = kwargs.pop('_event_actor')
        super().__init__(*args, **kwargs)

        self.fields['label'] = forms.CharField(
            label=_('Label'), required=False,
            widget=forms.TextInput(attrs={'readonly': 'readonly'})
        )
        self.fields['relationship_type'] = forms.ChoiceField(
            label=_('Relationship'),
            widget=forms.RadioSelect(), choices=self.RELATIONSHIP_CHOICES
        )

        self.sub_object = self.initial.get('sub_object')
        if self.sub_object:
            self.fields['label'].initial = str(self.sub_object)

            self.initial_relationship_type = self.get_relationship_type()

            self.fields['relationship_type'].initial = self.initial_relationship_type

    def get_new_relationship_instance(self):
        related_manager = getattr(
            self.initial.get('object'),
            self.initial['relationship_related_field']
        )
        main_field_name = related_manager.field.name

        return related_manager.model(
            **{
                main_field_name: self.initial.get('object'),
                self.initial['relationship_related_query_field']: self.initial.get('sub_object')
            }
        )

    def get_relationship_queryset(self):
        return getattr(
            self.initial.get('object'),
            self.initial['relationship_related_field']
        ).filter(
            **{
                self.initial['relationship_related_query_field']: self.initial.get('sub_object')
            }
        )

    def get_relationship_instance(self):
        relationship_queryset = self.get_relationship_queryset()
        if relationship_queryset.exists():
            return relationship_queryset.get()
        else:
            return self.get_new_relationship_instance()

    def save(self):
        if self.sub_object:
            if self.cleaned_data['relationship_type'] != self.initial_relationship_type:
                save_method = getattr(
                    self, 'save_relationship_{}'.format(
                        self.cleaned_data['relationship_type']
                    )
                )
                save_method()



# Generated by Django 2.2.20 on 2021-04-19 07:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('duplicates', '0009_auto_20210124_0738'),
    ]

    operations = [
        migrations.AlterField(
            model_name='duplicatebackendentry',
            name='documents',
            field=models.ManyToManyField(related_name='as_duplicate', to='documents.Document', verbose_name='Duplicated documents'),
        ),
    ]



from mayan.apps.documents.search import document_search
from mayan.apps.storage.utils import fs_cleanup, mkdtemp

from ..classes import SearchBackend
from ..settings import setting_backend_arguments


class SearchAPIViewTestMixin:
    def _request_search_view(self):
        query = {'q': self.test_document.label}
        return self.get(
            viewname='rest_api:search-view', kwargs={
                'search_model_name': document_search.get_full_name()
            }, query=query
        )

    def _request_advanced_search_view(self):
        query = {'document_type__label': self.test_document.document_type.label}

        return self.get(
            viewname='rest_api:advanced-search-view', kwargs={
                'search_model_name': document_search.get_full_name()
            }, query=query
        )


class SearchTestMixin:
    def setUp(self):
        self.old_value = setting_backend_arguments.value
        super().setUp()
        setting_backend_arguments.set(
            value={'index_path': mkdtemp()}
        )
        self.search_backend = SearchBackend.get_instance()

    def tearDown(self):
        fs_cleanup(
            filename=setting_backend_arguments.value['index_path']
        )
        setting_backend_arguments.set(value=self.old_value)
        super().tearDown()


class SearchToolsViewTestMixin:
    def _request_search_backend_reindex_view(self):
        return self.post(viewname='search:search_backend_reindex')


class SearchViewTestMixin:
    def _request_search_results_view(self, data, kwargs=None, query=None):
        return self.get(
            viewname='search:results', kwargs=kwargs, data=data, query=query
        )



from mayan.apps.document_states.permissions import permission_workflow_template_edit
from mayan.apps.document_states.tests.mixins.workflow_template_mixins import WorkflowTemplateTestMixin
from mayan.apps.document_states.tests.mixins.workflow_template_state_mixins import WorkflowTemplateStateActionViewTestMixin
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..models import MetadataType
from ..workflow_actions import (
    DocumentMetadataAddAction, DocumentMetadataEditAction,
    DocumentMetadataRemoveAction
)
from ..permissions import (
    permission_document_metadata_add, permission_document_metadata_edit,
    permission_document_metadata_remove
)

from .literals import (
    DOCUMENT_METADATA_ADD_ACTION_CLASS_PATH,
    DOCUMENT_METADATA_EDIT_ACTION_CLASS_PATH,
    DOCUMENT_METADATA_REMOVE_ACTION_CLASS_PATH, TEST_METADATA_VALUE
)
from .mixins import DocumentMetadataMixin, MetadataTypeTestMixin


class DocumentMetadataActionTestCase(
    DocumentMetadataMixin, MetadataTypeTestMixin
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_metadata_type()
        self.test_document_type.metadata.create(
            metadata_type=self.test_metadata_type
        )

    def test_document_metadata_add_action(self):
        action = DocumentMetadataAddAction(
            form_data={'metadata_types': MetadataType.objects.all()}
        )

        metadata_count = self.test_document.metadata.count()
        action.execute(context={'document': self.test_document})

        self.assertEqual(
            self.test_document.metadata.count(), metadata_count + 1
        )
        self.assertTrue(
            self.test_document.metadata.filter(
                metadata_type=self.test_metadata_type
            ).exists()
        )

    def test_document_metadata_edit_action(self):
        self._create_test_document_metadata()

        action = DocumentMetadataEditAction(
            form_data={
                'metadata_type': self.test_metadata_type.pk,
                'value': TEST_METADATA_VALUE
            }
        )

        metadata_value = self.test_document.metadata.first().value
        action.execute(context={'document': self.test_document})

        self.assertNotEqual(
            metadata_value, self.test_document.metadata.first().value
        )

    def test_document_metadata_remove_action(self):
        self._create_test_document_metadata()

        action = DocumentMetadataRemoveAction(
            form_data={'metadata_types': MetadataType.objects.all()}
        )

        metadata_count = self.test_document.metadata.count()
        action.execute(context={'document': self.test_document})

        self.assertEqual(
            self.test_document.metadata.count(), metadata_count - 1
        )
        self.assertFalse(
            self.test_document.metadata.filter(
                metadata_type=self.test_metadata_type
            ).exists()
        )


class DocumentMetadataActionViewTestCase(
    DocumentMetadataMixin, MetadataTypeTestMixin,
    WorkflowTemplateStateActionViewTestMixin, WorkflowTemplateTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self._create_test_metadata_type()
        self.test_document_type.metadata.create(
            metadata_type=self.test_metadata_type
        )
        self.test_workflow_template.document_types.add(
            self.test_document_type
        )

    def test_document_metadata_add_action_create_view(self):
        self.grant_access(
            obj=self.test_workflow_template, permission=permission_workflow_template_edit
        )
        self.grant_access(
            obj=self.test_metadata_type,
            permission=permission_document_metadata_add
        )

        response = self._request_test_workflow_template_state_action_create_post_view(
            class_path=DOCUMENT_METADATA_ADD_ACTION_CLASS_PATH, extra_data={
                'metadata_types': self.test_metadata_type.pk
            }
        )
        self.assertEqual(response.status_code, 302)

    def test_document_metadata_edit_action_create_view(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self.test_workflow_template, permission=permission_workflow_template_edit
        )
        self.grant_access(
            obj=self.test_metadata_type,
            permission=permission_document_metadata_edit
        )

        response = self._request_test_workflow_template_state_action_create_post_view(
            class_path=DOCUMENT_METADATA_EDIT_ACTION_CLASS_PATH, extra_data={
                'metadata_type': self.test_metadata_type.pk,
                'value': TEST_METADATA_VALUE
            }
        )
        self.assertEqual(response.status_code, 302)

    def test_document_metadata_remove_action_create_view(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self.test_workflow_template, permission=permission_workflow_template_edit
        )
        self.grant_access(
            obj=self.test_metadata_type,
            permission=permission_document_metadata_remove
        )

        response = self._request_test_workflow_template_state_action_create_post_view(
            class_path=DOCUMENT_METADATA_REMOVE_ACTION_CLASS_PATH, extra_data={
                'metadata_types': self.test_metadata_type.pk
            }
        )
        self.assertEqual(response.status_code, 302)



from django.utils.encoding import force_text

from rest_framework import status

from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import (
    event_cabinet_created, event_cabinet_edited, event_cabinet_document_added,
    event_cabinet_document_removed
)
from ..models import Cabinet
from ..permissions import (
    permission_cabinet_add_document, permission_cabinet_create,
    permission_cabinet_delete, permission_cabinet_edit,
    permission_cabinet_remove_document, permission_cabinet_view
)

from .mixins import (
    CabinetAPIViewTestMixin, CabinetDocumentAPIViewTestMixin,
    CabinetTestMixin, DocumentCabinetAPIViewTestMixin
)


class CabinetAPITestCase(
    CabinetAPIViewTestMixin, CabinetTestMixin, BaseAPITestCase
):
    def test_cabinet_create_api_view_no_permission(self):
        test_cabinet_count = Cabinet.objects.count()

        self._clear_events()

        response = self._request_test_cabinet_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(Cabinet.objects.count(), test_cabinet_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_create_api_view_with_permission(self):
        test_cabinet_count = Cabinet.objects.count()

        self.grant_permission(permission=permission_cabinet_create)

        self._clear_events()

        response = self._request_test_cabinet_create_api_view()

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data['id'], self.test_cabinet.pk)
        self.assertEqual(response.data['label'], self.test_cabinet.label)

        self.assertEqual(Cabinet.objects.count(), test_cabinet_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_cabinet)
        self.assertEqual(events[0].verb, event_cabinet_created.id)

    def test_cabinet_delete_api_view_no_permissions(self):
        self._create_test_cabinet()

        test_cabinet_count = Cabinet.objects.count()

        self._clear_events()

        response = self._request_test_cabinet_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Cabinet.objects.count(), test_cabinet_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_delete_api_view_with_access(self):
        self._create_test_cabinet()

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_delete
        )

        test_cabinet_count = Cabinet.objects.count()

        self._clear_events()

        response = self._request_test_cabinet_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(Cabinet.objects.count(), test_cabinet_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_edit_api_patch_view_no_pemission(self):
        self._create_test_cabinet()

        cabinet_label = self.test_cabinet.label

        self._clear_events()

        response = self._request_test_cabinet_edit_api_patch_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(cabinet_label, self.test_cabinet.label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_edit_api_patch_view_with_access(self):
        self._create_test_cabinet()

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_edit
        )

        cabinet_label = self.test_cabinet.label

        self._clear_events()

        response = self._request_test_cabinet_edit_api_patch_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_cabinet.refresh_from_db()
        self.assertNotEqual(cabinet_label, self.test_cabinet.label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_cabinet)
        self.assertEqual(events[0].verb, event_cabinet_edited.id)

    def test_cabinet_edit_api_put_view_no_pemission(self):
        self._create_test_cabinet()

        cabinet_label = self.test_cabinet.label

        self._clear_events()

        response = self._request_test_cabinet_edit_api_put_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(cabinet_label, self.test_cabinet.label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_edit_api_put_view_with_access(self):
        self._create_test_cabinet()

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_edit
        )

        cabinet_label = self.test_cabinet.label

        self._clear_events()

        response = self._request_test_cabinet_edit_api_put_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_cabinet.refresh_from_db()
        self.assertNotEqual(cabinet_label, self.test_cabinet.label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_cabinet)
        self.assertEqual(events[0].verb, event_cabinet_edited.id)

    def test_cabinet_list_api_view_no_permission(self):
        self._create_test_cabinet()

        self._clear_events()

        response = self._request_test_cabinet_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_list_api_view_with_access(self):
        self._create_test_cabinet()

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )

        self._clear_events()

        response = self._request_test_cabinet_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['label'], self.test_cabinet.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class CabinetDocumentAPITestCase(
    CabinetDocumentAPIViewTestMixin, CabinetTestMixin, DocumentTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_cabinet()

    def test_cabinet_document_add_api_view_no_permission(self):
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_add_document
        )

        test_cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_cabinet_document_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self.test_cabinet.documents.count(), test_cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_add_api_view_with_cabinet_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_add_document
        )

        test_cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_cabinet_document_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            self.test_cabinet.documents.count(), test_cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_add_api_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_add_document
        )

        test_cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_cabinet_document_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self.test_cabinet.documents.count(), test_cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_add_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_add_document
        )

        test_cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_cabinet_document_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            self.test_cabinet.documents.count(),
            test_cabinet_document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_cabinet)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document)
        self.assertEqual(events[0].verb, event_cabinet_document_added.id)

    def test_cabinet_trashed_document_add_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_add_document
        )

        test_cabinet_document_count = self.test_cabinet.documents.count()

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_cabinet_document_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            self.test_cabinet.documents.count(),
            test_cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_list_api_view_no_permission(self):
        self.test_cabinet.documents.add(self.test_document)

        self._clear_events()

        response = self._request_test_cabinet_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_list_api_view_with_cabinet_access(self):
        self.test_cabinet.documents.add(self.test_document)

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )

        self._clear_events()

        response = self._request_test_cabinet_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_list_api_view_with_document_access(self):
        self.test_cabinet.documents.add(self.test_document)

        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_view
        )

        self._clear_events()

        response = self._request_test_cabinet_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_list_api_view_with_full_access(self):
        self.test_cabinet.documents.add(self.test_document)

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_cabinet_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['results'][0]['uuid'],
            force_text(s=self.test_document.uuid)
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_trashed_document_list_api_view_with_full_access(self):
        self.test_cabinet.documents.add(self.test_document)

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_cabinet_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_remove_api_view_no_permission(self):
        self.test_cabinet.documents.add(self.test_document)

        test_cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_cabinet_document_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self.test_cabinet.documents.count(), test_cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_remove_api_view_with_cabinet_access(self):
        self.test_cabinet.documents.add(self.test_document)

        self.grant_access(
            obj=self.test_cabinet,
            permission=permission_cabinet_remove_document
        )

        test_cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_cabinet_document_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            self.test_cabinet.documents.count(), test_cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_remove_api_view_with_document_access(self):
        self.test_cabinet.documents.add(self.test_document)

        self.grant_access(
            obj=self.test_document,
            permission=permission_cabinet_remove_document
        )

        test_cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_cabinet_document_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self.test_cabinet.documents.count(), test_cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_remove_api_view_with_full_access(self):
        self.test_cabinet.documents.add(self.test_document)

        self.grant_access(
            obj=self.test_cabinet,
            permission=permission_cabinet_remove_document
        )
        self.grant_access(
            obj=self.test_document,
            permission=permission_cabinet_remove_document
        )

        test_cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_cabinet_document_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            self.test_cabinet.documents.count(),
            test_cabinet_document_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_cabinet)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document)
        self.assertEqual(events[0].verb, event_cabinet_document_removed.id)

    def test_cabinet_trashed_document_remove_api_view_with_full_access(self):
        self.test_cabinet.documents.add(self.test_document)

        self.grant_access(
            obj=self.test_cabinet,
            permission=permission_cabinet_remove_document
        )
        self.grant_access(
            obj=self.test_document,
            permission=permission_cabinet_remove_document
        )

        test_cabinet_document_count = self.test_cabinet.documents.count()

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_cabinet_document_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            self.test_cabinet.documents.count(), test_cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentCabinetAPITestCase(
    CabinetAPIViewTestMixin, CabinetTestMixin,
    DocumentCabinetAPIViewTestMixin, DocumentTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_cabinet()
        self.test_cabinet.documents.add(self.test_document)

    def test_document_cabinet_list_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_cabinet_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_cabinet_list_api_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_view,
        )

        self._clear_events()

        response = self._request_test_document_cabinet_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_cabinet_list_api_view_with_cabinet_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view,
        )

        self._clear_events()

        response = self._request_test_document_cabinet_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_cabinet_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_view,
        )
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view,
        )

        self._clear_events()

        response = self._request_test_document_cabinet_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['count'], Cabinet.objects.all().count()
        )
        self.assertEqual(
            response.data['results'][0]['id'], self.test_cabinet.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_cabinet_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_view,
        )
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view,
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_cabinet_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import json
import unittest

from django.db import connection, models

from rest_framework import serializers, status
from rest_framework.reverse import reverse

from mayan.apps.testing.tests.base import GenericViewTestCase

from .. import generics

from .base import BaseAPITestCase
from .literals import TEST_OBJECT_LABEL, TEST_OBJECT_LABEL_EDITED
from .mixins import RESTAPIViewTestMixin


class RESTAPIViewTestCase(RESTAPIViewTestMixin, GenericViewTestCase):
    def test_browser_api_view(self):
        response = self._request_test_browser_api_view()
        self.assertEqual(response.status_code, 200)

    @unittest.skipIf(connection.vendor != 'sqlite', 'Skip for known Django issues #15802 and #27074')
    def test_redoc_ui_view(self):
        response = self._request_test_redoc_ui_view()
        self.assertEqual(response.status_code, 200)

    @unittest.skipIf(connection.vendor != 'sqlite', 'Skip for known Django issues #15802 and #27074')
    def test_swagger_ui_view(self):
        response = self._request_test_swagger_ui_view()
        self.assertEqual(response.status_code, 200)

    def test_swagger_no_ui_json_view(self):
        self.expected_content_types = ('application/json; charset=utf-8',)

        response = self._request_test_swagger_no_ui_json_view()
        self.assertEqual(response.status_code, 200)

    def test_swagger_no_ui_yaml_view(self):
        self.expected_content_types = ('application/yaml; charset=utf-8',)

        response = self._request_test_swagger_no_ui_yaml_view()
        self.assertEqual(response.status_code, 200)


class BatchAPIRequestViewTestCase(BaseAPITestCase):
    def setUp(self):
        super().setUp()

        self._create_test_object(
            create_test_permission=True, fields={
                'label': models.CharField(max_length=32, unique=True)
            }, instance_kwargs={
                'label': TEST_OBJECT_LABEL
            }
        )

        class TestModelSerializer(serializers.ModelSerializer):
            class Meta:
                fields = ('id', 'label')
                model = self.TestModel

        def _test_view_factory():
            class TestView(generics.ListCreateAPIView):
                mayan_object_permissions = {
                    'GET': (self.test_permission,)
                }
                mayan_view_permissions = {
                    'POST': (self.test_permission,)
                }
                queryset = self.TestModel.objects.all()
                serializer_class = TestModelSerializer

            return TestView.as_view()

        self.add_test_view(
            test_object=self.test_object, test_view_factory=_test_view_factory,
            test_view_url=r'^test-view-url/$'
        )
        self._test_model_list_api_view_name = self._test_view_name

        def _test_view_factory():
            TestModel = self.TestModel

            class TestView(generics.RetrieveUpdateDestroyAPIView):
                lookup_url_kwarg = 'test_object_id'
                mayan_object_permissions = {
                    'DELETE': (self.test_permission,),
                    'GET': (self.test_permission,),
                    'PATCH': (self.test_permission,),
                    'PUT': (self.test_permission,)
                }
                queryset = TestModel.objects.all()
                serializer_class = TestModelSerializer

            return TestView.as_view()

        self.add_test_view(
            test_object=self.test_object, test_view_factory=_test_view_factory,
            test_view_url=r'^test-view-url/(?P<test_object_id>\d+)/$'
        )
        self._test_model_detail_api_view_name = self._test_view_name

    def _request_batch_api_request_api_view(self, requests):
        return self.post(
            viewname='rest_api:batchrequest-create', data={
                'requests': requests
            }
        )

    def test_create_batch_api_request(self):
        self.grant_permission(permission=self.test_permission)

        requests = [
            {
                'body': {'label': TEST_OBJECT_LABEL},
                'method': 'POST',
                'name': 'test_request',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_list_api_view_name
                    )
                )
            }
        ]

        self.test_object.delete()

        test_model_count = self.TestModel.objects.count()

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)

        self.assertEqual(
            response.data['results'][0]['status_code'], status.HTTP_201_CREATED
        )

        self.assertEqual(self.TestModel.objects.count(), test_model_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_delete_batch_api_request(self):
        self.grant_access(
            obj=self.test_object, permission=self.test_permission
        )

        requests = [
            {
                'method': 'DELETE',
                'name': 'test_request',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_detail_api_view_name
                    ), kwargs={'test_object_id': self.test_object.pk}
                )
            }
        ]

        test_model_count = self.TestModel.objects.count()

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)

        self.assertEqual(
            response.data['results'][0]['status_code'], status.HTTP_204_NO_CONTENT
        )

        self.assertEqual(self.TestModel.objects.count(), test_model_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_edit_via_patch_batch_api_request(self):
        self.grant_access(
            obj=self.test_object, permission=self.test_permission
        )

        requests = [
            {
                'body': {'label': TEST_OBJECT_LABEL_EDITED},
                'method': 'PATCH',
                'name': 'test_request',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_detail_api_view_name
                    ), kwargs={'test_object_id': self.test_object.pk}
                )
            }
        ]

        test_model_label = self.test_object.label

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)

        self.assertEqual(
            response.data['results'][0]['status_code'], status.HTTP_200_OK
        )

        self.test_object.refresh_from_db()
        self.assertNotEqual(self.test_object.label, test_model_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_edit_via_put_batch_api_request(self):
        self.grant_access(
            obj=self.test_object, permission=self.test_permission
        )

        requests = [
            {
                'body': {'label': TEST_OBJECT_LABEL_EDITED},
                'method': 'PUT',
                'name': 'test_request',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_detail_api_view_name
                    ), kwargs={'test_object_id': self.test_object.pk}
                )
            }
        ]

        test_model_label = self.test_object.label

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)

        self.assertEqual(
            response.data['results'][0]['status_code'], status.HTTP_200_OK
        )

        self.test_object.refresh_from_db()
        self.assertNotEqual(self.test_object.label, test_model_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_list_get_batch_api_request(self):
        self.grant_access(
            obj=self.test_object, permission=self.test_permission
        )

        requests = [
            {
                'name': 'test_request',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_list_api_view_name
                    )
                )
            }
        ]

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)

        self.assertEqual(
            response.data['results'][0]['status_code'], status.HTTP_200_OK
        )
        self.assertEqual(response.data['results'][0]['data']['count'], 1)
        self.assertEqual(
            response.data['results'][0]['data']['results'][0]['id'],
            self.test_object.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_mass_edit_get_batch_api_request(self):
        self.grant_access(
            obj=self.test_object, permission=self.test_permission
        )

        requests = [
            {
                'name': 'test_object_list',
                'url': reverse(
                    viewname='rest_api:{}'.format(
                        self._test_model_list_api_view_name
                    )
                )
            },
            {
                'body': {'label': TEST_OBJECT_LABEL_EDITED},
                'iterables': ['test_object_list.data.results'],
                'method': 'PATCH',
                'name': 'test_object_edit',
                'url': '{}{{{{ iterables.0.id }}}}/'.format(
                    reverse(
                        viewname='rest_api:{}'.format(
                            self._test_model_list_api_view_name
                        )
                    )
                )
            }
        ]

        test_model_label = self.test_object.label

        self._clear_events()

        response = self._request_batch_api_request_api_view(
            requests=json.dumps(obj=requests)
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 2)

        self.assertEqual(
            response.data['results'][0]['status_code'], status.HTTP_200_OK
        )
        self.assertEqual(response.data['results'][0]['data']['count'], 1)
        self.assertEqual(
            response.data['results'][0]['data']['results'][0]['id'],
            self.test_object.pk
        )

        self.assertEqual(
            response.data['results'][1]['status_code'], status.HTTP_200_OK
        )

        self.test_object.refresh_from_db()
        self.assertNotEqual(self.test_object.label, test_model_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging

from furl import furl

from django.conf import settings
from django.core import mail
from django.db import models
from django.urls import reverse
from django.utils.html import strip_tags
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import BackendModelMixin
from mayan.apps.templating.classes import Template

from .classes import NullBackend
from .events import event_email_sent
from .managers import UserMailerManager
from .utils import split_recipient_list

logger = logging.getLogger(name=__name__)


class UserMailer(BackendModelMixin, models.Model):
    """
    This model is used to create mailing profiles that can be used from inside
    the system. These profiles differ from the system mailing profile in that
    they can be created at runtime and can be assigned ACLs to restrict
    their use.
    """
    _backend_model_null_backend = NullBackend

    label = models.CharField(
        help_text=_('A short text describing the mailing profile.'),
        max_length=128, unique=True, verbose_name=_('Label')
    )
    default = models.BooleanField(
        default=True, help_text=_(
            'If default, this mailing profile will be pre-selected on the '
            'document mailing form.'
        ), verbose_name=_('Default')
    )
    enabled = models.BooleanField(default=True, verbose_name=_('Enabled'))
    backend_path = models.CharField(
        max_length=128,
        help_text=_('The dotted Python path to the backend class.'),
        verbose_name=_('Backend path')
    )
    backend_data = models.TextField(
        blank=True, verbose_name=_('Backend data')
    )

    objects = UserMailerManager()

    class Meta:
        ordering = ('label',)
        verbose_name = _('Mailing profile')
        verbose_name_plural = _('Mailing profiles')

    def __str__(self):
        return self.label

    def get_absolute_url(self):
        return reverse(viewname='mailer:user_mailer_list')

    def get_connection(self):
        """
        Establishes a reusable connection to the server by loading the
        backend, initializing it, and the using the backend instance to get
        a connection.
        """
        return mail.get_connection(
            backend=self.get_backend().class_path, **self.get_backend_data()
        )

    def natural_key(self):
        return (self.label,)

    def save(self, *args, **kwargs):
        if self.default:
            UserMailer.objects.select_for_update().exclude(pk=self.pk).update(
                default=False
            )

        return super().save(*args, **kwargs)

    def send(
        self, to, attachments=None, subject='', body='', cc=None, bcc=None,
        reply_to=None, _event_action_object=None, _user=None
    ):
        """
        Send a simple email. There is no document or template knowledge.
        attachments is a list of dictionaries with the keys:
        filename, content, and mimetype.
        """
        recipient_list = split_recipient_list(recipients=[to])

        if cc:
            cc_list = split_recipient_list(recipients=[cc])
        else:
            cc_list = None

        if bcc:
            bcc_list = split_recipient_list(recipients=[bcc])
        else:
            bcc_list = None

        if reply_to:
            reply_to_list = split_recipient_list(recipients=[reply_to])
        else:
            reply_to_list = None

        backend_data = self.get_backend_data()

        with self.get_connection() as connection:
            email_message = mail.EmailMultiAlternatives(
                body=strip_tags(body), connection=connection,
                from_email=backend_data.get('from'), subject=subject,
                to=recipient_list, cc=cc_list, bcc=bcc_list,
                reply_to=reply_to_list
            )

            for attachment in attachments or ():
                email_message.attach(
                    filename=attachment['filename'],
                    content=attachment['content'],
                    mimetype=attachment['mimetype']
                )

            email_message.attach_alternative(body, 'text/html')

        try:
            email_message.send()

        except Exception as exception:
            self.error_log.create(
                text='{}; {}'.format(
                    exception.__class__.__name__, exception
                )
            )
        else:
            self.error_log.all().delete()

            event_email_sent.commit(
                actor=_user, action_object=_event_action_object,
                target=self
            )

    def send_object(
        self, obj, to, as_attachment=False, body='', cc=None,
        content_function_dotted_path=None, bcc=None,
        mime_type_function_dotted_path=None,
        object_name=None, organization_installation_url='',
        reply_to=None, subject='', _user=None
    ):
        """
        Send an object file using this user mailing profile.
        """
        context_dictionary = {
            'link': furl(organization_installation_url).join(
                obj.get_absolute_url()
            ).tostr(),
            'object': obj,
            'object_name': object_name
        }

        body_template = Template(template_string=body)
        body_html_content = body_template.render(
            context=context_dictionary
        )

        subject_template = Template(template_string=subject)
        subject_text = strip_tags(
            subject_template.render(context=context_dictionary)
        )

        attachments = []
        if as_attachment:
            if not content_function_dotted_path:
                raise ValueError(
                    'Must provide `content_function_dotted_path` '
                    'to allow sending the object as an attachment.'
                )

            if not mime_type_function_dotted_path:
                raise ValueError(
                    'Must provide `mime_type_function_dotted_path` to '
                    'allow sending the object as an attachment.'
                )

            content_function = import_string(
                dotted_path=content_function_dotted_path
            )

            mime_type_function = import_string(
                dotted_path=mime_type_function_dotted_path
            )
            mime_type = mime_type_function(obj=obj)

            with content_function(obj=obj) as file_object:
                attachments.append(
                    {
                        'content': file_object.read(),
                        'filename': str(obj),
                        'mimetype': mime_type
                    }
                )

        return self.send(
            attachments=attachments, cc=cc, bcc=bcc, body=body_html_content,
            reply_to=reply_to, subject=subject_text, to=to,
            _event_action_object=obj, _user=_user
        )

    def test(self, to):
        """
        Send a test message to make sure the mailing profile settings are
        correct.
        """
        try:
            self.send(subject=_('Test email from Mayan EDMS'), to=to)
        except Exception as exception:
            self.error_log.create(
                text='{}; {}'.format(
                    exception.__class__.__name__, exception
                )
            )
            if settings.DEBUG:
                raise
        else:
            self.error_log.all().delete()



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import DEFAULT_VIEWS_PAGINATE_BY

namespace = SettingNamespace(label=_('Views'), name='views')

setting_paginate_by = namespace.add_setting(
    default=DEFAULT_VIEWS_PAGINATE_BY, global_name='VIEWS_PAGINATE_BY',
    help_text=_(
        'The number objects that will be displayed per page.'
    )
)



import operator

DEFAULT_RESULTS_LIMIT = 100
DEFAULT_SEARCH_BACKEND = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'
DEFAULT_SEARCH_BACKEND_ARGUMENTS = {}
DEFAULT_SEARCH_DISABLE_SIMPLE_SEARCH = False
DEFAULT_SEARCH_MATCH_ALL_DEFAULT_VALUE = 'false'
DEFAULT_SEARCH_RESULTS_LIMIT = 100

DEFAULT_SCOPE_ID = '0'

DELIMITER = '_'

SEARCH_MODEL_NAME_KWARG = 'search_model_name'

SCOPE_MARKER = '__'
SCOPE_MATCH_ALL = 'match_all'
SCOPE_OPERATOR_MARKER = 'operator'
SCOPE_RESULT_MAKER = 'result'

SCOPE_OPERATOR_AND = 'AND'
SCOPE_OPERATOR_OR = 'OR'
SCOPE_OPERATOR_NOT = 'NOT'

SCOPE_OPERATOR_CHOICES = {
    SCOPE_OPERATOR_AND: operator.and_,
    SCOPE_OPERATOR_OR: operator.or_,
    SCOPE_OPERATOR_NOT: operator.not_
}



import logging

from django.apps import apps

from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app

from .classes import SearchBackend, SearchModel

logger = logging.getLogger(name=__name__)


@app.task(
    bind=True, ignore_result=True, max_retries=None, retry_backoff=True
)
def task_deindex_instance(self, app_label, model_name, object_id):
    logger.info('Executing')

    Model = apps.get_model(app_label=app_label, model_name=model_name)
    instance = Model._meta.default_manager.get(pk=object_id)

    try:
        SearchBackend.get_instance().deindex_instance(instance=instance)
    except LockError as exception:
        raise self.retry(exc=exception)

    logger.info('Finished')


@app.task(
    bind=True, ignore_result=True, max_retries=None, retry_backoff=True
)
def task_index_search_model(self, search_model_full_name):
    search_model = SearchModel.get(name=search_model_full_name)

    for instance in search_model.model._meta.default_manager.all():
        task_index_instance.apply_async(
            kwargs={
                'app_label': instance._meta.app_label,
                'model_name': instance._meta.model_name,
                'object_id': instance.pk
            }
        )


@app.task(
    bind=True, ignore_result=True, max_retries=None, retry_backoff=True
)
def task_index_instance(self, app_label, model_name, object_id):
    logger.info('Executing')

    try:
        Model = apps.get_model(app_label=app_label, model_name=model_name)
    except LookupError:
        """
        The app or model does not exists anymore. Non fatal, just exit
        the task.
        """
    else:
        try:
            instance = Model._meta.default_manager.get(pk=object_id)
        except Model.DoesNotExist:
            """
            This is not fatal, the task is triggered on the post_save
            signal and the instance might still not be ready to access.
            """
        else:
            try:
                SearchBackend.get_instance().index_instance(instance=instance)
            except LockError as exception:
                raise self.retry(exc=exception)

    logger.info('Finished')



from django.apps import apps
from django.urls import reverse_lazy

from mayan.apps.dashboards.classes import DashboardWidgetList

from .links import link_current_user_events


class DashboardWidgetUserEvents(DashboardWidgetList):
    columns = ('event_type', 'target')
    icon = link_current_user_events.icon
    label = link_current_user_events.text
    link = reverse_lazy(
        viewname=link_current_user_events.view
    )

    def get_object_list(self):
        Action = apps.get_model(
            app_label='actstream', model_name='Action'
        )

        return Action.objects.actor(obj=self.request.user)



from django.contrib import admin

from .models import (
    DocumentFilePageContent, DocumentFileParseError
)


@admin.register(DocumentFilePageContent)
class DocumentFilePageContentAdmin(admin.ModelAdmin):
    list_display = ('document_file_page',)


@admin.register(DocumentFileParseError)
class DocumentFileParseErrorAdmin(admin.ModelAdmin):
    list_display = ('document_file', 'datetime_submitted')
    readonly_fields = ('document_file', 'datetime_submitted', 'result')



from datetime import timedelta

from django.db import models
from django.utils.timezone import now

from mayan.apps.events.classes import EventModelRegistry, ModelEventType

from .events import (
    event_download_file_created, event_download_file_deleted
)
from .literals import (
    INTERVAL_DOWNLOAD_FILE_EXPIRATION, INTERVAL_SHARED_UPLOAD_STALE
)


class DownloadFileManager(models.Manager):
    def stale(self):
        return self.filter(
            datetime__lt=now() - timedelta(
                seconds=INTERVAL_DOWNLOAD_FILE_EXPIRATION
            )
        )

    def register_content_object(self, model):
        EventModelRegistry.register(model=model)

        ModelEventType.register(
            model=model, event_types=(
                event_download_file_created, event_download_file_deleted
            )
        )


class SharedUploadedFileManager(models.Manager):
    def stale(self):
        return self.filter(
            datetime__lt=now() - timedelta(
                seconds=INTERVAL_SHARED_UPLOAD_STALE
            )
        )



from rest_framework import status

from mayan.apps.documents.tests.base import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_smart_link_edited
from ..models import SmartLinkCondition
from ..permissions import (
    permission_smart_link_edit, permission_smart_link_view
)

from .literals import (
    TEST_SMART_LINK_CONDITION_EXPRESSION,
    TEST_SMART_LINK_CONDITION_EXPRESSION_EDITED,
    TEST_SMART_LINK_CONDITION_OPERATOR
)
from .mixins import SmartLinkConditionAPIViewTestMixin, SmartLinkTestMixin


class SmartLinkConditionAPIViewTestCase(
    DocumentTestMixin, SmartLinkTestMixin,
    SmartLinkConditionAPIViewTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_smart_link(add_test_document_type=True)

    def test_smart_link_condition_create_api_view_no_permission(self):
        self._clear_events()

        response = self._request_smart_link_condition_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_condition_create_api_view_with_access(self):
        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_smart_link_condition_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        smart_link_condition = SmartLinkCondition.objects.first()
        self.assertEqual(response.data['id'], smart_link_condition.pk)
        self.assertEqual(
            response.data['operator'], TEST_SMART_LINK_CONDITION_OPERATOR
        )

        self.assertEqual(SmartLinkCondition.objects.count(), 1)
        self.assertEqual(
            smart_link_condition.operator, TEST_SMART_LINK_CONDITION_OPERATOR
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self.test_smart_link_condition
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_smart_link)
        self.assertEqual(events[0].verb, event_smart_link_edited.id)

    def test_smart_link_condition_delete_api_view_no_permission(self):
        self._create_test_smart_link_condition()

        self._clear_events()

        response = self._request_smart_link_condition_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(SmartLinkCondition.objects.count(), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_condition_delete_api_view_with_access(self):
        self._create_test_smart_link_condition()

        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_smart_link_condition_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(SmartLinkCondition.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_smart_link)
        self.assertEqual(events[0].verb, event_smart_link_edited.id)

    def test_smart_link_condition_detail_api_view_no_permission(self):
        self._create_test_smart_link_condition()

        self._clear_events()

        response = self._request_smart_link_condition_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_condition_detail_api_view_with_access(self):
        self._create_test_smart_link_condition()

        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_view
        )

        self._clear_events()

        response = self._request_smart_link_condition_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['operator'], TEST_SMART_LINK_CONDITION_OPERATOR
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_condition_edit_via_patch_api_view_no_permission(self):
        self._create_test_smart_link_condition()

        self._clear_events()

        response = self._request_smart_link_condition_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_smart_link_condition.refresh_from_db()
        self.assertEqual(
            self.test_smart_link_condition.expression,
            TEST_SMART_LINK_CONDITION_EXPRESSION
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_condition_edit_via_patch_api_view_with_access(self):
        self._create_test_smart_link_condition()

        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_smart_link_condition_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_smart_link_condition.refresh_from_db()
        self.assertEqual(
            self.test_smart_link_condition.expression,
            TEST_SMART_LINK_CONDITION_EXPRESSION_EDITED
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self.test_smart_link_condition
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_smart_link)
        self.assertEqual(events[0].verb, event_smart_link_edited.id)

    def test_smart_link_condition_edit_via_put_api_view_no_permission(self):
        self._create_test_smart_link_condition()

        self._clear_events()

        response = self._request_smart_link_condition_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_smart_link_condition.refresh_from_db()
        self.assertEqual(
            self.test_smart_link_condition.expression,
            TEST_SMART_LINK_CONDITION_EXPRESSION
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_condition_edit_via_put_api_view_with_access(self):
        self._create_test_smart_link_condition()

        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_smart_link_condition_edit_via_put_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_smart_link_condition.refresh_from_db()
        self.assertEqual(
            self.test_smart_link_condition.expression,
            TEST_SMART_LINK_CONDITION_EXPRESSION_EDITED
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object, self.test_smart_link_condition
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_smart_link)
        self.assertEqual(events[0].verb, event_smart_link_edited.id)

    def test_smart_link_condition_list_api_view_no_permission(self):
        self._create_test_smart_link_condition()

        self._clear_events()

        response = self._request_smart_link_condition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_condition_list_api_view_with_access(self):
        self._create_test_smart_link_condition()

        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_view
        )

        self._clear_events()

        response = self._request_smart_link_condition_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['id'],
            self.test_smart_link_condition.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging

from rest_framework import status
from rest_framework.response import Response

from mayan.apps.converter.api_view_mixins import APIImageViewMixin
from mayan.apps.rest_api import generics
from mayan.apps.storage.models import SharedUploadedFile
from mayan.apps.views.generics import DownloadViewMixin

from ..permissions import (
    permission_document_file_delete, permission_document_file_download,
    permission_document_file_edit, permission_document_file_new,
    permission_document_file_view
)
from ..serializers.document_file_serializers import (
    DocumentFileSerializer, DocumentFilePageSerializer
)
from ..tasks import task_document_file_upload

from .mixins import (
    ParentObjectDocumentAPIViewMixin, ParentObjectDocumentFileAPIViewMixin
)

logger = logging.getLogger(name=__name__)


class APIDocumentFileListView(
    ParentObjectDocumentAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Return a list of the selected document's files.
    post: Create a new document file.
    """
    ordering_fields = ('comment', 'encoding', 'id', 'mime_type')
    serializer_class = DocumentFileSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        return Response(status=status.HTTP_202_ACCEPTED)

    def perform_create(self, serializer):
        shared_uploaded_file = SharedUploadedFile.objects.create(
            file=serializer.validated_data['file_new']
        )

        task_document_file_upload.apply_async(
            kwargs={
                'action': serializer.validated_data['action'],
                'comment': serializer.validated_data.get('comment', ''),
                'document_id': self.get_document(
                    permission=permission_document_file_new
                ).pk,
                'filename': serializer.validated_data.get('filename', ''),
                'shared_uploaded_file_id': shared_uploaded_file.pk,
                'user_id': self.request.user.pk
            }
        )

    def get_queryset(self):
        return self.get_document(
            permission=permission_document_file_view
        ).files.all()


class APIDocumentFileDetailView(
    ParentObjectDocumentAPIViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected document file.
    get: Returns the selected document file details.
    """
    lookup_url_kwarg = 'document_file_id'
    mayan_object_permissions = {
        'DELETE': (permission_document_file_delete,),
        'GET': (permission_document_file_view,),
        'PATCH': (permission_document_file_edit,),
        'PUT': (permission_document_file_edit,),
    }
    serializer_class = DocumentFileSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_queryset(self):
        return self.get_document().files.all()


class APIDocumentFileDownloadView(
    DownloadViewMixin, ParentObjectDocumentAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Download a document file.
    """
    lookup_url_kwarg = 'document_file_id'
    mayan_object_permissions = {
        'GET': (permission_document_file_download,),
    }

    def get_download_file_object(self):
        instance = self.get_object()
        instance._event_actor = self.request.user
        return instance.get_download_file_object()

    def get_download_filename(self):
        return self.get_object().filename

    def get_serializer(self, *args, **kwargs):
        return None

    def get_serializer_class(self):
        return None

    def get_queryset(self):
        return self.get_document().files.all()

    def retrieve(self, request, *args, **kwargs):
        return self.render_to_response()


# Document file page


class APIDocumentFilePageDetailView(
    ParentObjectDocumentFileAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Returns the selected document page details.
    """
    lookup_url_kwarg = 'document_file_page_id'
    serializer_class = DocumentFilePageSerializer
    mayan_object_permissions = {
        'GET': (permission_document_file_view,),
    }

    def get_queryset(self):
        return self.get_document_file().pages.all()


class APIDocumentFilePageImageView(
    APIImageViewMixin, ParentObjectDocumentFileAPIViewMixin,
    generics.RetrieveAPIView
):
    """
    get: Returns an image representation of the selected document.
    """
    lookup_url_kwarg = 'document_file_page_id'
    mayan_object_permissions = {
        'GET': (permission_document_file_view,),
    }

    def get_queryset(self):
        return self.get_document_file().pages.all()


class APIDocumentFilePageListView(
    ParentObjectDocumentFileAPIViewMixin, generics.ListAPIView
):
    serializer_class = DocumentFilePageSerializer

    def get_queryset(self):
        return self.get_document_file(
            permission=permission_document_file_view
        ).pages.all()



import logging

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import OperationalError

from mayan.celery import app

from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError

from .literals import DEFAULT_SOURCES_LOCK_EXPIRE

logger = logging.getLogger(name=__name__)


@app.task(ignore_result=True)
def task_source_process_document(source_id, dry_run=False):
    Source = apps.get_model(
        app_label='sources', model_name='Source'
    )

    lock_id = 'task_source_process_document-%d' % source_id
    try:
        logger.debug('trying to acquire lock: %s', lock_id)
        lock = LockingBackend.get_backend().acquire_lock(
            name=lock_id, timeout=DEFAULT_SOURCES_LOCK_EXPIRE
        )
    except LockError:
        logger.debug('unable to obtain lock: %s' % lock_id)
    else:
        logger.debug('acquired lock: %s', lock_id)

        try:
            source = Source.objects.get(pk=source_id)
            if source.enabled or dry_run:
                source.get_backend_instance().process_documents(dry_run=dry_run)
        except Exception as exception:
            logger.error(
                'Error processing source id: %s; %s', source_id, exception,
                exc_info=True
            )
            if settings.DEBUG:
                raise
        finally:
            lock.release()


@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_process_document_upload(
    self, document_type_id, shared_uploaded_file_id, source_id,
    callback_kwargs=None, description=None, expand=False, label=None,
    language=None, user_id=None
):
    try:
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        SharedUploadedFile = apps.get_model(
            app_label='storage', model_name='SharedUploadedFile'
        )
        Source = apps.get_model(
            app_label='sources', model_name='Source'
        )

        document_type = DocumentType.objects.get(pk=document_type_id)
        shared_uploaded_file = SharedUploadedFile.objects.get(
            pk=shared_uploaded_file_id
        )
        source = Source.objects.get(pk=source_id)

        if not label:
            label = shared_uploaded_file.filename

        if user_id:
            user = get_user_model().objects.get(pk=user_id)
        else:
            user = None
    except OperationalError as exception:
        logger.warning(
            'Operational error during attempt to load data to handle source '
            'upload: %s. Retrying.', exception
        )
        raise self.retry(exc=exception)
    else:
        with shared_uploaded_file.open() as file_object:
            source.handle_file_object_upload(
                callback_kwargs=callback_kwargs, document_type=document_type,
                description=description, expand=expand,
                file_object=file_object, label=label, language=language,
                user=user
            )

        shared_uploaded_file.delete()



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Menu

from .icons import icon_menu_about, icon_menu_user

menu_about = Menu(
    icon=icon_menu_about, label=_('System'), name='about'
)
menu_facet = Menu(label=_('Facet'), name='facet')
menu_list_facet = Menu(label=_('Facet'), name='list facet')
menu_main = Menu(name='main')
menu_multi_item = Menu(name='multi item')
menu_object = Menu(label=_('Actions'), name='object')
menu_related = Menu(label=_('Related'), name='related')
menu_secondary = Menu(label=_('Secondary'), name='secondary')
menu_setup = Menu(name='setup')
menu_return = Menu(label=_('Return'), name='return')
menu_tools = Menu(name='tools')
menu_topbar = Menu(name='topbar')
menu_user = Menu(
    icon=icon_menu_user, name='user', label=_('User')
)



from actstream.models import Action
from rest_framework import serializers
from rest_framework.reverse import reverse

from mayan.apps.common.serializers import ContentTypeSerializer
from mayan.apps.rest_api.fields import DynamicSerializerField
from mayan.apps.user_management.serializers import UserSerializer

from .classes import EventType
from .models import Notification, StoredEventType


class EventTypeNamespaceSerializer(serializers.Serializer):
    label = serializers.CharField()
    name = serializers.CharField()
    url = serializers.SerializerMethodField()

    event_types_url = serializers.HyperlinkedIdentityField(
        lookup_field='name',
        view_name='rest_api:event-type-namespace-event-type-list',
    )

    def get_url(self, instance):
        return reverse(
            viewname='rest_api:event-type-namespace-detail', kwargs={
                'name': instance.name
            }, request=self.context['request'], format=self.context['format']
        )


class EventTypeSerializer(serializers.Serializer):
    label = serializers.CharField()
    name = serializers.CharField()
    id = serializers.CharField()
    event_type_namespace_url = serializers.SerializerMethodField()

    def get_event_type_namespace_url(self, instance):
        return reverse(
            viewname='rest_api:event-type-namespace-detail', kwargs={
                'name': instance.namespace.name
            }, request=self.context['request'], format=self.context['format']
        )

    def to_representation(self, instance):
        if isinstance(instance, EventType):
            return super().to_representation(instance=instance)
        elif isinstance(instance, StoredEventType):
            return super().to_representation(instance=instance.get_class())
        elif isinstance(instance, str):
            return super().to_representation(
                instance=EventType.get(name=instance)
            )


class EventSerializer(serializers.ModelSerializer):
    actor = DynamicSerializerField(read_only=True)
    target = DynamicSerializerField(read_only=True)
    actor_content_type = ContentTypeSerializer(read_only=True)
    target_content_type = ContentTypeSerializer(read_only=True)
    verb = EventTypeSerializer(read_only=True)

    class Meta:
        exclude = (
            'action_object_content_type', 'action_object_object_id'
        )
        model = Action


class NotificationSerializer(serializers.ModelSerializer):
    user = UserSerializer(read_only=True)
    action = EventSerializer(read_only=True)

    class Meta:
        fields = ('action', 'read', 'user')
        model = Notification



from .parsers import MetadataParser
from .validators import MetadataValidator

DEFAULT_METADATA_AVAILABLE_VALIDATORS = MetadataValidator.get_import_paths()
DEFAULT_METADATA_AVAILABLE_PARSERS = MetadataParser.get_import_paths()



from datetime import timedelta

from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_c

from .literals import CHECK_EXPIRED_CHECK_OUTS_INTERVAL

queue_checkouts_periodic = CeleryQueue(
    label=_('Checkouts periodic'), name='checkouts_periodic', transient=True,
    worker=worker_c
)

queue_checkouts_periodic.add_task_type(
    label=_('Check expired checkouts'),
    name='task_check_expired_check_outs',
    dotted_path='mayan.apps.checkouts.tasks.task_check_expired_check_outs',
    schedule=timedelta(
        seconds=CHECK_EXPIRED_CHECK_OUTS_INTERVAL
    ),
)



from mayan.apps.common.tests.mixins import ObjectCopyTestMixin
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from .mixins import TagTestMixin


class TagCopyTestCase(
    DocumentTestMixin, TagTestMixin, ObjectCopyTestMixin, BaseTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_tag(add_test_document=True)
        self.test_object = self.test_tag



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0012_auto_20170205_0743'),
    ]

    operations = [
        migrations.AddField(
            model_name='sanescanner',
            name='adf_mode',
            field=models.CharField(
                blank=True, choices=[
                    ('simplex', 'Simples'), ('duplex', 'Duplex')
                ], default='simplex',
                help_text='Selects the document feeder mode '
                '(simplex/duplex). If this option is not supported by '
                'your scanner, leave it blank.', max_length=16,
                verbose_name='ADF mode'
            ),
        ),
        migrations.AddField(
            model_name='sanescanner',
            name='source',
            field=models.CharField(
                blank=True, choices=[
                    ('flatbed', 'Flatbed'),
                    ('document-feeder', 'Document feeder')
                ], default='flatbed', help_text='Selects the scan source '
                '(such as a document-feeder). If this option is not '
                'supported by your scanner, leave it blank.', max_length=16,
                verbose_name='Paper source'
            ),
        ),
        migrations.AlterField(
            model_name='sanescanner',
            name='mode',
            field=models.CharField(
                blank=True, choices=[
                    ('lineart', 'Lineart'), ('monochrome', 'Monochrome'),
                    ('color', 'Color')
                ], default='color', help_text='Selects the scan mode '
                '(e.g., lineart, monochrome, or color). If this option is '
                'not supported by your scanner, leave it blank.',
                max_length=16, verbose_name='Mode'
            ),
        ),
        migrations.AlterField(
            model_name='sanescanner',
            name='resolution',
            field=models.PositiveIntegerField(
                blank=True, help_text='Sets the resolution of the scanned '
                'image in DPI (dots per inch). Typical value is 200. If '
                'this option is not supported by your scanner, leave it '
                'blank.', verbose_name='Resolution'
            ),
        ),
    ]
